'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _promise = require('babel-runtime/core-js/promise');

var _promise2 = _interopRequireDefault(_promise);

var _assign = require('babel-runtime/core-js/object/assign');

var _assign2 = _interopRequireDefault(_assign);

var _stringify = require('babel-runtime/core-js/json/stringify');

var _stringify2 = _interopRequireDefault(_stringify);

exports.default = reusePromise;
exports.clear = clear;
exports.decorator = decorator;

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

// TODO compare key with shallow equal, not JSON
// TODO equality check function

var pendingPromisesMap = function () {
  var FN_INDEX = 0;
  var KEY_INDEX = 1;
  var VALUE_INDEX = 2;

  var a = [];

  function serializeKey(key) {
    return (0, _stringify2.default)(key);
  }

  function is(item, fn, serializedKey) {
    if (item[FN_INDEX] == fn && item[KEY_INDEX] == serializedKey) {
      return true;
    }
    return false;
  }

  function getIndex(fn, serializedKey) {
    for (var i = 0, l = a.length; i < l; i++) {
      var item = a[i];

      if (is(item, fn, serializedKey)) {
        return i;
      }
    }

    return -1;
  }

  function get(fn, key) {
    var serializedKey = serializeKey(key);
    var i = getIndex(fn, serializedKey);

    if (a[i]) {
      return a[i][VALUE_INDEX];
    }

    return null;
  }

  function set(fn, key, value) {
    var serializedKey = serializeKey(key);
    var i = getIndex(fn, serializedKey);

    if (i > -1) {
      a[i][VALUE_INDEX] = value;
    } else {
      var obj = [];
      obj[FN_INDEX] = fn;
      obj[KEY_INDEX] = serializedKey;
      obj[VALUE_INDEX] = value;
      a.push(obj);
    }

    return value;
  }

  function del(fn, key) {
    if (key === undefined) {
      // delete all of function
      for (var i = a.length - 1; i >= 0; i--) {
        var item = a[i];
        if (item[FN_INDEX] == fn) {
          a.splice(i, 1);
        }
      }
    } else {
      // delete single entry by function and key
      var serializedKey = serializeKey(key);
      var _i = getIndex(fn, serializedKey);
      if (_i > -1) {
        a.splice(_i, 1);
      }
    }
  }

  function clear() {
    a.length = 0;
  }

  return {
    get: get,
    set: set,
    del: del,
    clear: clear
  };
}();

function reusePromise(origFn) {
  var options = arguments.length <= 1 || arguments[1] === undefined ? {} : arguments[1];

  options = (0, _assign2.default)({ memoize: false }, options);

  var wrappedFn = function wrappedFn() {
    for (var _len = arguments.length, args = Array(_len), _key = 0; _key < _len; _key++) {
      args[_key] = arguments[_key];
    }

    var key = args;
    var memoizeValueKey = ['memoized', key];

    var pendingPromise = pendingPromisesMap.get(origFn, key);

    if (pendingPromise) {
      return pendingPromise;
    }

    if (options.memoize) {
      var prevValue = pendingPromisesMap.get(origFn, memoizeValueKey);

      if (prevValue) {
        return _promise2.default.resolve(prevValue);
      }
    }

    var forgetPromise = function forgetPromise() {
      pendingPromisesMap.del(origFn, key);
    };

    var origPromise = origFn.apply(this, args);

    var promise = origPromise.then(function (value) {
      if (options.memoize) {
        pendingPromisesMap.set(origFn, memoizeValueKey, value);
      }

      forgetPromise();
      return value;
    }, function (err) {
      forgetPromise();
      throw err;
    });
    pendingPromisesMap.set(origFn, key, promise);

    return promise;
  };

  wrappedFn.__reusePromise__origFn = origFn;
  wrappedFn.__reusePromise__clear = function () {
    reusePromise.clear(origFn);
  };

  return wrappedFn;
}

function clear() {
  var fn = arguments.length <= 0 || arguments[0] === undefined ? undefined : arguments[0];

  if (fn == undefined) {
    pendingPromisesMap.clear();
  } else {
    pendingPromisesMap.del(fn.__reusePromise__origFn || fn);
  }
}

function decorator() {
  var options = arguments.length <= 0 || arguments[0] === undefined ? {} : arguments[0];

  return function (target, name, descriptor) {
    descriptor.value = reusePromise(descriptor.value, options);
    return descriptor;
  };
}

decorator.clear = clear;
reusePromise.clear = clear;